@@ -624,6 +624,12 @@ mb_wm_handle_property_notify (XPropertyEvent          *xev,
     flag = MBWM_WINDOW_PROP_NET_PID;
   else if (xev->atom == wm->atoms[MBWM_ATOM_HILDON_STACKING_LAYER])
     flag = MBWM_WINDOW_PROP_HILDON_STACKING;
+  else if (xev->atom == wm->atoms[MBWM_ATOM_HILDON_PORTRAIT_MODE_SUPPORT])
+    flag = MBWM_WINDOW_PROP_PORTRAIT;
+  else if (xev->atom == wm->atoms[MBWM_ATOM_HILDON_PORTRAIT_MODE_REQUEST])
+    flag = MBWM_WINDOW_PROP_PORTRAIT;
+  else if (xev->atom == wm->atoms[MBWM_ATOM_HILDON_LIVE_DESKTOP_BACKGROUND])
+    flag = MBWM_WINDOW_PROP_LIVE_BACKGROUND;
 
   if (flag)
     mb_wm_client_window_sync_properties (client->window, flag);
@@ -685,7 +691,7 @@ mb_wm_handle_root_config_notify (XConfigureEvent *xev,
   /* Bastard hack part 2: now that the screen is reconfigured
    * map and activate the client which caused us to rotate. */
   for (c = wm->stack_top; c; c = c->stacked_below)
-    if (!mb_wm_client_is_map_confirmed (c) && c->window->portrait_on_map)
+    if (!mb_wm_client_is_map_confirmed (c) && c->window->portrait_requested > 0)
       {
         mb_wm_activate_client (wm, c);
         break;
@@ -1063,12 +1069,64 @@ stack_sync_to_display (MBWindowManager *wm)
   mb_wm_util_async_untrap_x_errors();
 }
 
+extern gboolean hd_dbus_tklock_on;
+
+static void
+mb_wm_focus_client_as_stacked (MBWindowManager *wm,
+                               MBWindowManagerClient *except_client)
+{
+  MBWindowManagerClient *next = NULL;
+
+  if (wm->stack_top)
+    {
+      MBWindowManagerClient *c;
+
+      mb_wm_stack_enumerate_reverse (wm, c)
+	{
+	  if (c != except_client &&
+	      mb_wm_client_want_focus (c) &&
+	      mb_wm_client_is_visible (c) &&
+              /* do not assign focus to Home applets automatically */
+              c->window->net_type !=
+                wm->atoms[MBWM_ATOM_HILDON_WM_WINDOW_TYPE_HOME_APPLET])
+	    {
+	      next = c;
+	      break;
+	    }
+	  if (mb_wm_client_covers_screen (c))
+	    {
+	      /* anything below this is necessarily
+	       * invisible
+	       */
+	      break;
+	    }
+	}
+    }
+
+  wm->focused_client = NULL;
+
+  if (hd_dbus_tklock_on)
+    {
+      /* don't give focus back to any already mapped window
+       * because the touch screen is still locked and we don't want
+       * the application to wake up */
+      return;
+    }
+
+  if (next)
+    {
+      mb_wm_focus_client (wm, next);
+    }
+}
+
 void
 mb_wm_sync (MBWindowManager *wm)
 {
   /* Sync all changes to display */
   MBWindowManagerClient *client = NULL;
+#ifdef TIME_MB_WM_SYNC
   GTimer *timer = g_timer_new();
+#endif
   MBWM_MARK();
   MBWM_TRACE ();
 
@@ -1126,8 +1184,16 @@ mb_wm_sync (MBWindowManager *wm)
   XFlush(wm->xdpy);
   wm->sync_type = 0;
 
+  if (wm->focus_after_stacking)
+    {
+      wm->focus_after_stacking = False;
+      mb_wm_focus_client_as_stacked (wm, NULL);
+    }
+
+#ifdef TIME_MB_WM_SYNC
   g_debug("mb_wm_sync: %f", g_timer_elapsed(timer,0));
   g_timer_destroy(timer);
+#endif
 }
 
 static void
@@ -1250,20 +1316,7 @@ mb_wm_manage_client (MBWindowManager       *wm,
   if (MB_WM_CLIENT_CLIENT_TYPE (client) != MBWMClientTypeOverride)
     sync_flags |= MBWMSyncStacking;
 
-#if ENABLE_COMPOSITE
-  if (mb_wm_comp_mgr_enabled (wm->comp_mgr) && !client->cm_client)
-    mb_wm_comp_mgr_register_client (wm->comp_mgr, client);
-#endif
-
-  if (!activate || MB_WM_CLIENT_CLIENT_TYPE (client) == MBWMClientTypeDesktop)
-    mb_wm_client_show (client);
-  else if (!mb_wm_client_wants_portrait(client))
-    mb_wm_activate_client (wm, client);
-  else
-    /* Leave it up to the desktop to do something and we'll show
-     * and activate the client when the screen size changes. */
-    mb_wm_object_signal_emit (MB_WM_OBJECT (wm),
-                              MBWindowManagerSignalPortraitForecast);
+  mb_wm_comp_mgr_register_client (wm->comp_mgr, client, activate);
 
   mb_wm_display_sync_queue (client->wmref, sync_flags);
 }
@@ -1423,12 +1476,21 @@ mb_wm_main_loop(MBWindowManager *wm)
 
 void
 mb_wm_get_display_geometry (MBWindowManager  *wm,
-			    MBGeometry       *geometry)
+			    MBGeometry       *geometry,
+                            Bool              use_layout_flag)
 {
   geometry->x = 0;
   geometry->y = 0;
-  geometry->width  = wm->xdpy_width;
-  geometry->height = wm->xdpy_height;
+  if (use_layout_flag && (wm->flags & MBWindowManagerFlagLayoutRotated))
+    {
+      geometry->width  = wm->xdpy_height;
+      geometry->height = wm->xdpy_width;
+    }
+  else
+    {
+      geometry->width  = wm->xdpy_width;
+      geometry->height = wm->xdpy_height;
+    }
 }
 
 void
@@ -1491,7 +1553,7 @@ mb_wm_manage_preexisting_wins (MBWindowManager* wm)
 	        * we can process map notification on the frame.
 	        */
 	       if (wm->comp_mgr && mb_wm_comp_mgr_enabled (wm->comp_mgr))
-		 mb_wm_comp_mgr_register_client (wm->comp_mgr, client);
+		 mb_wm_comp_mgr_register_client (wm->comp_mgr, client, False);
 #endif
 	       mb_wm_manage_client(wm, client, False);
 	     }
@@ -1500,6 +1562,8 @@ mb_wm_manage_preexisting_wins (MBWindowManager* wm)
 	 }
      }
 
+   wm->focus_after_stacking = True;
+
    XFree(wins);
 }
 
@@ -2060,6 +2124,40 @@ mb_wm_handle_show_desktop (MBWindowManager * wm, Bool show)
     }
 }
 
+static Bool
+timestamp_predicate (Display *display,
+                     XEvent  *xevent,
+                     XPointer arg)
+{
+  MBWindowManager *wm = (MBWindowManager *)arg;
+  Window xwindow = wm->root_win->xwindow;
+
+  if (xevent->type == PropertyNotify &&
+      xevent->xproperty.window == xwindow &&
+      xevent->xproperty.atom == wm->atoms[MBWM_ATOM_NET_WORKAREA])
+    return True;
+
+  return False;
+}
+
+Time
+mb_wm_get_server_time (MBWindowManager *wm)
+{
+  XEvent xevent;
+  long data = 0;
+
+  /* zero-length append to get timestamp in the PropertyNotify */
+  XChangeProperty (wm->xdpy, wm->root_win->xwindow,
+                   wm->atoms[MBWM_ATOM_NET_WORKAREA],
+                   XA_CARDINAL, 32, PropModeAppend,
+                   (unsigned char*)&data, 0);
+
+  XIfEvent (wm->xdpy, &xevent, timestamp_predicate, (XPointer)wm);
+
+  return xevent.xproperty.time;
+}
+
+
 static void
 mb_wm_set_layout (MBWindowManager *wm, MBWMLayout *layout)
 {
@@ -2068,12 +2166,11 @@ mb_wm_set_layout (MBWindowManager *wm, MBWMLayout *layout)
 }
 
 static void
-mb_wm_focus_client (MBWindowManager *wm, MBWindowManagerClient *c)
+mb_wm_focus_client (MBWindowManager *wm, MBWindowManagerClient *client)
 {
-  MBWindowManagerClient *client = c,
-    *last_focused_transient;
+  MBWindowManagerClient *last_focused_transient;
 
-  last_focused_transient = mb_wm_client_get_last_focused_transient (c);
+  last_focused_transient = mb_wm_client_get_last_focused_transient (client);
 
   /*
    * If the last focused transient for this client is modal, we try to focus
@@ -2122,14 +2219,15 @@ mb_wm_focus_client (MBWindowManager *wm, MBWindowManagerClient *c)
 
       mb_wm_client_realize (client);
       mb_wm_client_display_sync (client);
-    }
 
-  if (mb_wm_client_focus (client))
-    {
-      wm->focused_client = client;
+      /* focus what ever should be focused according to the stacking order,
+       * because this window could be e.g. behind the current application */
+      wm->focus_after_stacking = True;
       return;
     }
 
+  mb_wm_client_focus (client);
+
   /*
    * If the window is already destroyed we will get a notification about it
    * later.
@@ -2138,56 +2236,13 @@ mb_wm_focus_client (MBWindowManager *wm, MBWindowManagerClient *c)
   return;
 }
 
-extern gboolean hd_dbus_tklock_on;
-
 void
 mb_wm_unfocus_client (MBWindowManager *wm, MBWindowManagerClient *client)
 {
-  MBWindowManagerClient *next = NULL;
-
   if (client != wm->focused_client)
     return;
 
-  if (wm->stack_top)
-    {
-      MBWindowManagerClient *c;
-
-      mb_wm_stack_enumerate_reverse (wm, c)
-	{
-	  if (c != client &&
-	      mb_wm_client_want_focus (c) &&
-	      mb_wm_client_is_visible (c) &&
-              /* do not assign focus to Home applets automatically */
-              c->window->net_type !=
-                wm->atoms[MBWM_ATOM_HILDON_WM_WINDOW_TYPE_HOME_APPLET])
-	    {
-	      next = c;
-	      break;
-	    }
-	  if (mb_wm_client_covers_screen (c))
-	    {
-	      /* anything below this is necessarily
-	       * invisible
-	       */
-	      break;
-	    }
-	}
-    }
-
-  wm->focused_client = NULL;
-
-  if (hd_dbus_tklock_on)
-    {
-      /* don't give focus back to any already mapped window
-       * because the touch screen is still locked and we don't want
-       * the application to wake up */
-      return;
-    }
-
-  if (next)
-    {
-      mb_wm_focus_client (wm, next);
-    }
+  mb_wm_focus_client_as_stacked (wm, client);
 }
 
 void
